#include <cmath>
#include <limits>
#include <gsl/gsl_sf.h>
#include "integration.hh"

#include "commons.hh"
#include "QQbarPair.hh"
#include "math/special_functions.hh"


using namespace std;

namespace HELL {

  double ACTIVATE_AB = 1;
  double ACTIVATE_NAB = 1;

  // matrix element squared with 2 off-shell legs
  double Mel2_QQbar(double xi1, double xi2, double z, double phi, double z1, double z2, double pk2) { //, double pk1=0) { // pk2=pt*k2/Q^2
    double xi = xi1+xi2+2*sqrt(xi1*xi2)*cos(phi);
    double S = 1/z - xi;
    double T = xi2 + 2*sqrt(xi2*xi1)*cos(phi) - 2*pk2 - (1-z1)/z;
    //double T = 2*pk1 - xi1 - z2 /z;
    double U = 2*pk2 - xi2 - z1 /z;
    //cout << T << "  " << U << endl;
    double zero = S+T+U+xi1+xi2;
    if(fabs(zero)>1e-6) cout << "S+T+U+xi1+xi2 = "<< zero << endl;
    double Delta = (z1-z2)/z + xi1*z2 - xi2*z1 + (U+xi2-T-xi1)/2.;
    double M2ab  =  1/T/U - pow(1+z2*(1-z1)/z/T+z1*(1-z2)/z/U,2)/xi1/xi2;
    double M2nab = -1/T/U + (z1-z2)*(T-U)/S/T/U + 2*z/S + 2*(0.5+z2*(1-z1)/z/T-Delta/S)*(0.5+z1*(1-z2)/z/U+Delta/S)/xi1/xi2;
    double M2 = ( ACTIVATE_AB*M2ab/2./CA + ACTIVATE_NAB*M2nab/CF/4. );
    return M2/z/z;
  }
  // matrix element squared with 1 off-shell leg, averaged over phi
  double Mel2_QQbar_partialonshell(double xi1, double z, double z1, double z2, double pthat2) {
    double zxi = 1-z*xi1;
    //
    double M2ab  = (1-2*z2*(1-z2))/z1/(1-z1) - 2/xi1*pow((1-z1-z2)/z1/(1-z1),2)*(pthat2-z1*z2/z);
    double M2nab = -1/z1/(1-z1) + (1-2*z1)*(z2-z1)/z1/(1-z1)/zxi + 2/zxi
      + 2*z2/zxi/zxi*( 2*z2*(z1-z*xi1)/(1-z1) +1-z2*zxi/(1-z1) + (1-zxi/(1-z1))*(1-(1-z2)*zxi/z1) )
      + 2/zxi/zxi/xi1*( -2*pthat2*(1-z2*zxi/(1-z1))*(1-(1-z2)*zxi/z1)
			+z2*(z1+z2-1)/z*( 1-z2*zxi/(1-z1) + (1-zxi/(1-z1))*(1-(1-z2)*zxi/z1) ) );
    double M2 = ( ACTIVATE_AB*M2ab/2./CA + ACTIVATE_NAB*M2nab/CF/4. );
    return M2;
  }
  // matrix element squared with 1 off-shell leg, with phi dependence
  double Mel2_QQbar_partialonshell(double xi1, double phi, double z, double z1, double z2, double pthat2) {
    double zxi = 1-z*xi1;
    double cost = ((z1+z2-1)/z+xi1)/2/sqrt(xi1*pthat2);
    double sint = -sqrt(1-cost*cost); // which sign of sint?
    if(sint!=sint) sint = 0;
    double cosp = cos(phi);
    double costp = cosp*cost + sin(phi)*sint;
    //
    double M2ab  = 1/z1/(1-z1) - 4/xi1*pow(z2/(1-z1)*sqrt(xi1)*cosp + ((1-z2)/z1-z2/(1-z1))*sqrt(pthat2)*costp,2);
    double M2nab = -1/z1/(1-z1) + (1-2*z1)*(z2-z1)/z1/(1-z1)/zxi + 2/zxi
      + 8/xi1
      * ( (z2/zxi-z2/(1-z1))*sqrt(xi1)*cosp - (1/zxi-z2/(1-z1))*sqrt(pthat2)*costp )
      * ( -z2/zxi*sqrt(xi1)*cosp + (1/zxi-(1-z2)/z1)*sqrt(pthat2)*costp );
    double M2 = ( ACTIVATE_AB*M2ab/2./CA + ACTIVATE_NAB*M2nab/CF/4. );
    if(M2 != M2) {
      cout << M2 << "  " << xi1 << "  " << z << "  " << phi << "  " << z1 << "  " << z2 << "  " << pthat2 << "  " << sint << "  " << cost << endl;
      M2 = 0;
    }
    return M2;
  }
  // matrix element squared fully on-shell
  double Mel2_QQbar_onshell(double z, double z1, double pthat2) {
    double M2ab  = 1/z1/(1-z1) * (1 + 4*z*pthat2*(1-z*pthat2/z1/(1-z1))) - 2;
    double M2nab = 4*z1*(1-z1) - 2 - 8*z*pthat2*(1-z*pthat2/z1/(1-z1));
    double M2 = ( ACTIVATE_AB*M2ab/2./CA + ACTIVATE_NAB*M2nab/CF/4. );
    return M2;
  }
  double Mel2_QQbar_onshell_qqbarchannel(double z, double pthat2) {
    return CF/CA * (1-2*z*pthat2);
  }
  //
  double phiprime(double xi1, double xi2, double phi) { // angle between k1+k2 and k2
    double qthat2 = xi1+xi2+2*sqrt(xi1*xi2)*cos(phi);
    double cosphip = (sqrt(xi2)+sqrt(xi1)*cos(phi))/sqrt(qthat2);
    if(fabs(cosphip)-1>1e-8) cout << "Warning (phiprime), how can it be?" << endl;
    double phip = myacos(cosphip);
    if(sin(phi)<0) phip = 2*M_PI - phip; // corrects the angle if it's between pi and 2*pi [comes from the fact that sin(phip)\propto sin(phi)]
    return phip;
  }
  double phinonprime(double xi, double xi2, double phiprime) { // angle between k1 and k2
    double xi1 = xi+xi2-2*sqrt(xi*xi2)*cos(phiprime);
    double cosphi = (-sqrt(xi2)+sqrt(xi)*cos(phiprime))/sqrt(xi1);
    if(fabs(cosphi)-1>1e-8) cout << "Warning (phinonprime): how can it be?" << endl;
    double phi = myacos(cosphi);
    if(sin(phiprime)<0) phi = 2*M_PI - phi; // corrects the angle if it's between pi and 2*pi [comes from the fact that sin(phi)\propto sin(phip)]
    //if(phi!=phi) cout << (-sqrt(xi2)+sqrt(xi)*cos(phiprime))/sqrt(xi1) <<endl;
    //cout << setw(12) << phiprime << setw(12) << phi << setw(14) << sin(phiprime) << setw(14) << cosphi << setw(14) << xi1 << endl;  //////////////////////////
    return phi;
  }








  /////////////////////
  // Q-Qbar Pair
  /////////////////////


  // initialization
  QQbarPair::QQbarPair(double mQ, double muFrat) {
    _mQ     = mQ;
    _muFrat = muFrat;
    //areCoeffsSaved = false;
    //
  }
  //
  void   QQbarPair::SetKinematics(double x, double invM, double rapY, double pt) {
    _x = x;
    _invM=invM;
    _rapY=rapY;
    _pt=pt;
    if(invM<=2*_mQ) {
      cout << "Error: you are setting an invariant mass M=" << invM << " for the heavy quark pair which is smaller than the minimum allowed 2m=" << 2*_mQ << endl;
      abort();
    }
  }
  //
  // basic object dC/dphi
  double QQbarPair::dCoff_dphi(double xi1, double xi2, double phi, double omega, double beta) { // the last two are phase-space integration vars
    //if(_invM<2*_mQ) return 0; // violates theta function  // redundant
    double qthat2 = xi1+xi2+2*sqrt(xi1*xi2)*cos(phi);
    double z = 1./(1.+qthat2);
    double m2Q2 = _mQ*_mQ/_invM/_invM;
    //
    double absDeltahat = sqrt(1-4*m2Q2)*sin(beta)/2.;
    double z1 = 0.5-sqrt(0.25-m2Q2)*cos(beta);
    double z2 = z * ( 1 - (1-qthat2)*z1 + 2*sqrt(qthat2)*absDeltahat*cos(omega) );
    if(sqrt(0.25-m2Q2)-fabs(0.5-z2)<0) { cout << "eeeeeee" << endl; }
    if(z2<0 || z2>1) { cout << "ooooooo" << endl; }
    //
    double phip = phiprime(xi1,xi2,phi);
    double pk2 = z1*(xi2+sqrt(xi1*xi2)*cos(phi)) + sqrt(xi2)*absDeltahat*cos(omega+phip);
    //
    double M2 = Mel2_QQbar(xi1, xi2, z, phi, z1, z2, pk2);
    return M2 * sqrt(0.25-m2Q2)*sin(beta) /16./M_PI/M_PI /4./M_PI;
  }
  // its integrated version in the xi2->0 limit: int_0^2pi dphi dC/dphi(xi2=0)   (note: it's not an average, it's the full integral)
  double QQbarPair::Coff(double xi1, double omega, double beta) { // the last two are phase-space integration vars
    //if(_invM<2*_mQ) return 0; // violates theta function  // redundant
    double qthat2 = xi1;
    double z = 1./(1.+qthat2);
    double m2Q2 = _mQ*_mQ/_invM/_invM;
    //
    double absDeltahat = sqrt(1-4*m2Q2)*sin(beta)/2.;
    double z1 = 0.5-sqrt(0.25-m2Q2)*cos(beta);
    double z2 = z * ( 1 - (1-qthat2)*z1 + 2*sqrt(qthat2)*absDeltahat*cos(omega) );
    if(sqrt(0.25-m2Q2)-fabs(0.5-z2)<0) { cout << "eeeeeee" << endl; }
    if(z2<0 || z2>1) { cout << "ooooooo" << endl; }
    //
    double pthat2 = z1*z1*qthat2 + absDeltahat*absDeltahat + 2*z1*sqrt(qthat2)*absDeltahat*cos(omega);
    //
    double M2 = Mel2_QQbar_partialonshell(xi1, z, z1, z2, pthat2);
    return M2 * sqrt(0.25-m2Q2)*sin(beta) /16./M_PI/M_PI /2.;
  }

  // basic object dC/dphi, partially on-shell, namely the limit for xi2=0 of dC/dphi, averaged over phi
  double QQbarPair::dCoff_dphi_0(double xi1, double omega, double beta) { // the last two are phase-space integration vars
    return Coff(xi1, omega, beta) /2./M_PI;
    /*
    //if(_invM<2*_mQ) return 0; // violates theta function  // redundant
    double qthat2 = xi1;
    double z = 1./(1.+qthat2);
    double m2Q2 = _mQ*_mQ/_invM/_invM;
    //
    double absDeltahat = sqrt(1-4*m2Q2)*sin(beta)/2.;
    double z1 = 0.5-sqrt(0.25-m2Q2)*cos(beta);
    double z2 = z * ( 1 - (1-qthat2)*z1 + 2*sqrt(qthat2)*absDeltahat*cos(omega) );
    if(sqrt(0.25-m2Q2)-fabs(0.5-z2)<0) { cout << "eeeeeee" << endl; }
    if(z2<0 || z2>1) { cout << "ooooooo" << endl; }
    //
    double pthat2 = z1*z1*xi1 + absDeltahat*absDeltahat + 2*z1*sqrt(xi1)*absDeltahat*cos(omega);
    //
    double M2 = Mel2_QQbar_partialonshell(xi1, z, z1, z2, pthat2);
    return M2 * sqrt(0.25-m2Q2)*sin(beta) /16./M_PI/M_PI /4./M_PI;
    */
  }
  // basic object dC/dphi, partially on-shell, namely the limit for xi2=0 of dC/dphi, keeping phi dependence
  double QQbarPair::dCoff_dphi_0(double xi1, double phi, double omega, double beta) { // the last two are phase-space integration vars
    //if(_invM<2*_mQ) return 0; // violates theta function  // redundant
    double qthat2 = xi1;
    double z = 1./(1.+qthat2);
    double m2Q2 = _mQ*_mQ/_invM/_invM;
    //
    double absDeltahat = sqrt(1-4*m2Q2)*sin(beta)/2.;
    double z1 = 0.5-sqrt(0.25-m2Q2)*cos(beta);
    double z2 = z * ( 1 - (1-qthat2)*z1 + 2*sqrt(qthat2)*absDeltahat*cos(omega) );
    if(sqrt(0.25-m2Q2)-fabs(0.5-z2)<0) { cout << "eeeeeee" << endl; }
    if(z2<0 || z2>1) { cout << "ooooooo" << endl; }
    //
    double pthat2 = z1*z1*xi1 + absDeltahat*absDeltahat + 2*z1*sqrt(xi1)*absDeltahat*cos(omega);
    //
    double M2 = Mel2_QQbar_partialonshell(xi1, phi, z, z1, z2, pthat2);
    return M2 * sqrt(0.25-m2Q2)*sin(beta) /16./M_PI/M_PI /4./M_PI;
  }


  // on-shell version: int_0^2pi dphi dC/dphi(xi1=0,xi2=0)   (note: it's not an average, it's the full integral)
  double QQbarPair::Con(double beta) {
    //if(_invM<2*_mQ) return 0; // violates theta function  // redundant
    double m2Q2 = _mQ*_mQ/_invM/_invM;
    //
    double absDeltahat = sqrt(1-4*m2Q2)*sin(beta)/2.;
    double z1 = 0.5-sqrt(0.25-m2Q2)*cos(beta);
    double z2 = 1-z1;
    if(sqrt(0.25-m2Q2)-fabs(0.5-z2)<0) { cout << "eeeeeee" << endl; }
    if(z2<0 || z2>1) { cout << "ooooooo" << endl; }
    //
    double pthat2 = absDeltahat*absDeltahat;
    //
    double M2 = Mel2_QQbar_onshell(1, z1, pthat2); // correct, verified
    //double M2 = Mel2_QQbar_partialonshell(1e-9, 1, z1, z2, pthat2); // don't use it - limit not so smooth
    return M2 * sqrt(0.25-m2Q2)*sin(beta) /16./M_PI/M_PI /2.  *2*M_PI;
  }
  double QQbarPair::Con_qqbarchannel(double beta) {
    //if(_invM<2*_mQ) return 0; // violates theta function  // redundant
    double m2Q2 = _mQ*_mQ/_invM/_invM;
    //
    double absDeltahat = sqrt(1-4*m2Q2)*sin(beta)/2.;
    double z1 = 0.5-sqrt(0.25-m2Q2)*cos(beta);
    double z2 = 1-z1;
    if(sqrt(0.25-m2Q2)-fabs(0.5-z2)<0) { cout << "eeeeeee" << endl; }
    if(z2<0 || z2>1) { cout << "ooooooo" << endl; }
    //
    double pthat2 = absDeltahat*absDeltahat;
    //
    double M2 = Mel2_QQbar_onshell_qqbarchannel(1, pthat2);
    return M2 * sqrt(0.25-m2Q2)*sin(beta) /16./M_PI/M_PI /2.  *2*M_PI;
  }





  // triple differential
  double QQbarPair::diffXS_M_Y_pt(double xi1, double xi2, double intvar1, double intvar2) {
    //if(_invM<2*_mQ) return 0; // violates theta function
    double qthat2 = pow(_pt/_invM,2);
    if(exp(-2*fabs(_rapY))/_x < 1+qthat2) {
      cout << "Warning: if condition in action" << endl;
      return 0;
    }
    double cosphi = (qthat2-xi1-xi2)/2./sqrt(xi1*xi2);
    if(fabs(cosphi)>1.) {
      cout << "theta function in action: |cos(phi)|-1 = " << fabs(cosphi)-1 << "  xi1="<<xi1 << "  xi2="<<xi2 << "  qthat2="<<qthat2 << endl;
      return 0;
    }
    double omega = 2*M_PI*intvar1;
    //double beta1 = M_PI*intvar1;
    double beta2 = M_PI*intvar2;
    double phi = myacos(cosphi);
    double sqrtarg = 4*xi1*xi2-(qthat2-xi1-xi2)*(qthat2-xi1-xi2);
    if(sqrtarg<=0) {
      cout << "Warning: argument of sqrt is negative: " << sqrtarg << "  and cosphi = " << cosphi << endl;
      sqrtarg = 1e-20;
    }
    //
    double dCdphi = ( dCoff_dphi(xi1, xi2, phi, omega, beta2) + dCoff_dphi(xi1, xi2, 2*M_PI-phi, omega, beta2) ) * 2*M_PI*M_PI;
    double res = dCdphi / sqrt(sqrtarg) /(1+qthat2);
    if(std::isinf(res)) {
      cout << "Coff is infinite, and sqrtarg = " << sqrtarg << endl;
      res = std::numeric_limits<float>::max();
    }
    return res;
  }
  double QQbarPair::diffXS_M_Y_pt_SUB(double xi1, double intvar0, double intvar1, double intvar2, double &xi2p) {
    //if(_invM<2*_mQ) return 0; // violates theta function
    double qthat2 = pow(_pt/_invM,2);
    if(exp(-2*fabs(_rapY))/_x < 1+qthat2) {
      cout << "Warning: if condition in action" << endl;
      return 0;
    }
    //double cosphimax = qthat2/2./xi1-1.;
    double phimin = 0;//(cosphimax>1? 0 : myacos(cosphimax));
    double phi   = phimin + 2*(M_PI-phimin)*intvar2;
    double omega = 2*M_PI*intvar0;
    //double beta1 = M_PI*intvar1;
    double beta2 = M_PI*intvar1;
    double cosphi = cos(phi);
    //
    double res = 0;
    xi2p=1;
    if(xi1<qthat2/2./(1.+cosphi)) {
    //if(cosphi<cosphimax) {
      //cout << "siii" << endl;
      //cout << cosphimax << "  " << phimin << "  " << phi << "  " << qthat2/2./(1.+cosphi)-xi1 << endl;
      double sqrtaux = sqrt(qthat2-xi1*(1-cosphi*cosphi));
      double sqrtxi2p = -sqrt(xi1)*cosphi + sqrtaux;
      xi2p = sqrtxi2p*sqrtxi2p;
      if(xi2p<xi1) cout << "something is not working properly" << endl;
      res += dCoff_dphi(sqrtxi2p*sqrtxi2p, xi1, phi, omega, beta2) * sqrtxi2p / sqrtaux *(1.-phimin/M_PI); // correction to the jacobian due to the change of variable for phi for this piece only
      //res += sqrtxi2p*(1-xi1)*(1-sqrtxi2p*sqrtxi2p);
      // if(sub) {
      // 	//cout << "subtraction" << endl;
      // 	//res -= dCoff_dphi_0(qthat2, omega, beta2) / qthat2;
      // 	res -= dCoff_dphi_0(qthat2, phi, omega, beta2) / qthat2;  // the on-shell limit is smooth for each phi, and not only after integrating over phi
      // }
    } else {
      cout << "noooo   xi1=" <<xi1 << "  xi1max=" << qthat2/2./(1.+cosphi) << endl;
    }
    // if(xi1<qthat2 && sub) {
    //   //cout << "subtraction" << endl;
    //   //res -= dCoff_dphi_0(qthat2, omega, beta2) / qthat2;
    //   res -= dCoff_dphi_0(qthat2, phi, omega, beta2) / qthat2;  // the on-shell limit is smooth for each phi, and not only after integrating over phi
    // }
    //
    if(res != res) {
      //cout << res << "  " << xi1 << "  " << intvar0 << "  " << intvar1 << "  " << intvar2 << endl;
      res = 0;
    }
    return res * 4*M_PI*M_PI*M_PI /(1+qthat2);
  }
  double QQbarPair::diffXS_M_Y_pt_SUBx(double intvar0, double intvar1, double intvar2) {
    //if(_invM<2*_mQ) return 0; // violates theta function
    double qthat2 = pow(_pt/_invM,2);
    if(exp(-2*fabs(_rapY))/_x < 1+qthat2) {
      cout << "Warning: if condition in action" << endl;
      return 0;
    }
    double phimin = 0;
    double phi   = phimin + 2*(M_PI-phimin)*intvar2;
    double omega = 2*M_PI*intvar0;
    //double beta1 = M_PI*intvar1;
    double beta2 = M_PI*intvar1;
    //
    double res = dCoff_dphi_0(qthat2, phi, omega, beta2);  // the on-shell limit is smooth for each phi, and not only after integrating over phi
    //
    if(res != res) {
      //cout << res << "  " << xi1 << "  " << intvar0 << "  " << intvar1 << "  " << intvar2 << endl;
      res = 0;
    }
    return res * 4*M_PI*M_PI*M_PI /(1+qthat2);
  }
  /*
  double QQbarPair::diffXS_M_Y_pt_SUBc(double intvar1, double intvar2) {
    //if(_invM<2*_mQ) return 0; // violates theta function
    double qthat2 = pow(_pt/_invM,2);
    if(exp(-2*fabs(_rapY))/_x < 1+qthat2) {
      cout << "Warning: if condition in action" << endl;
      return 0;
    }
    if(qthat2>1) return 0;
    double omega = 2*M_PI*intvar1;
    //double beta1 = M_PI*intvar1;
    double beta2 = M_PI*intvar2;
    //
    return 2*log(qthat2)/qthat2 * dCoff_dphi_0(qthat2, omega, beta2) * 2*M_PI*M_PI /(1+qthat2);
  }
  */
  /*
  double QQbarPair::diffXS_M_Y_pt_v2(double xi2, double phip, double intvar1, double intvar2) {
    //if(_invM<2*_mQ) return 0; // violates theta function
    double qthat2 = pow(_pt/_invM,2);
    if(1/_x < 1+qthat2) {
      cout << "Warning: if condition in action" << endl;
      return 0;
    }
    double omega = 2*M_PI*intvar1;
    //double beta1 = M_PI*intvar1;
    double beta2 = M_PI*intvar2;
    double xi1 = qthat2 + xi2 - 2*sqrt(qthat2*xi2)*cos(phip);
    double phi = phinonprime(qthat2, xi2, phip);
    //
    double dCdphi = dCoff_dphi(xi1, xi2, phi, omega, beta2) * 2*M_PI*M_PI;
    return dCdphi /(1+qthat2);
  }
  */
  double QQbarPair::diffXS_M_Y_pt_Aux(double intvar1, double intvar2) {
    //if(_invM<2*_mQ) return 0; // violates theta function
    double qthat2 = pow(_pt/_invM,2);
    if(1/_x < 1+qthat2) {
      cout << "Warning: if condition in action" << endl;
      return 0;
    }
    double omega = 2*M_PI*intvar1;
    double beta2 =   M_PI*intvar2;
    //
    double C = Coff(qthat2, omega, beta2) * 2*M_PI*M_PI;
    return C /(1+qthat2);
  }






  // double differential M Y
  double QQbarPair::intphi(double intvar, double xi1, double xi2, double &jac) {
    double cosphibarmax = (exp(-2.*fabs(_rapY))/_x-1.-xi1-xi2)/2./sqrt(xi1*xi2);
    double phibar = 0;
    if     (cosphibarmax<-1) phibar = M_PI;
    else if(cosphibarmax< 1) phibar = myacos(cosphibarmax);
    if(phibar<0)    cout << phibar << endl;
    if(phibar>M_PI) cout << phibar << endl;
    jac = 2*(M_PI-phibar);
    double phi = phibar + jac*intvar;
    //jac = 2*M_PI;
    //phi = jac*intvar;
    return phi;
  }
  double QQbarPair::diffXS_M_Y(double xi1, double xi2, double intvar1, double intvar2, double phi) {
    //if(_invM<2*_mQ) return 0; // violates theta function
    double omega = 2*M_PI*intvar1;
    //double beta1 =   M_PI*intvar1;
    double beta2 =   M_PI*intvar2;
    if(exp(-2*fabs(_rapY))/_x-1<pow(sqrt(xi1)-sqrt(xi2),2)) {
      //cout << "Warning: first if condition in action. x=" << _x << " y=" << _rapY << " xi1=" << xi1 << " xi2=" << xi2 << endl;
      return 0;
    }
    double cosphi = cos(phi);
    double xi = xi1+xi2+2*sqrt(xi1*xi2)*cosphi;
    if(exp(-2.*fabs(_rapY))/_x-1. < xi) {
      cout << "Warning: second if condition in action:  " << exp(-2*fabs(_rapY))/_x-1 << "  " << xi << "  " << xi1 << "  " << xi2 << "  " << cosphi << endl;
      return 0;
    }
    double dCdphi = dCoff_dphi(xi1, xi2, phi, omega, beta2) * 2*M_PI*M_PI;
    return dCdphi /(1+xi);
  }
  double QQbarPair::diffXS_M_Y_Aux(double intvar1, double intvar2, double xi) {
    //if(_invM<2*_mQ) return 0; // violates theta function
    //if(_rapY>0) return 0;
    //double z = _x*exp(-2*_rapY);
    //double xi = 1/z-1;
    //if(_x > exp(2*_rapY)) {
    //  cout << "Warning: if condition in action" << endl;
    //  return 0;
    //}
    double omega = 2*M_PI*intvar1;
    double beta2 =   M_PI*intvar2;
    //
    double C = Coff(xi, omega, beta2) * 2*M_PI*M_PI;
    return C;
  }
  double QQbarPair::diffXS_M_Y_LO(double intvar1, double intvar2, double intvar3) {
    //if(_invM<2*_mQ) return 0; // violates theta function
    //
    double omega = 2*M_PI*intvar1;
    double beta2 =   M_PI*intvar2;
    //double phi =   2*M_PI*intvar3;
    //
    //double C = Con(beta2) * M_PI;
    double C = Coff(1e-12, omega, beta2) * 2*M_PI*M_PI;
    //double C = dCoff_dphi_0(1e-12, phi, omega, beta2) * 2*M_PI*2*M_PI*M_PI;
    //cout << omega << "  " << beta2 << "  " << C << endl;
    return C;
  }
  double QQbarPair::diffXS_M_Y_LO_qqbarchannel(double intvar1) {
    //if(_invM<2*_mQ) return 0; // violates theta function
    double beta2 =   M_PI*intvar1;
    double C = Con_qqbarchannel(beta2) * M_PI;
    return C;
  }












  /////////////////////
  // Q-Qbar Single Quark
  /////////////////////


  // initialization
  QQbarSingleQuark::QQbarSingleQuark(double mQ, double muFrat) {
    _mQ     = mQ;
    _muFrat = muFrat;
    //areCoeffsSaved = false;
    //
  }
  //
  void   QQbarSingleQuark::SetKinematics(double x, double rapY, double pt) {
    _x = x;
    _rapY=rapY;
    _pt=pt;
  }





  // double differential
  double QQbarSingleQuark::diffXS_Y_pt(double xi1, double xi2, double z, double y, double intvar1) {
    //if(_invM<2*_mQ) return 0; // violates theta function
    double phi = 2*M_PI*intvar1;
    double pthat  = _pt/_mQ;
    double qthat2 = xi1+xi2+2*sqrt(xi1*xi2)*cos(phi);
    double sqrtcoshy = sqrt(1+pthat*pthat)*cosh(y);
    double oozmax = pow(sqrtcoshy+sqrt(sqrtcoshy*sqrtcoshy-pthat*pthat+pow(pthat+sqrt(qthat2),2)),2);
    double oozmin = pow(sqrtcoshy+sqrt(sqrtcoshy*sqrtcoshy-pthat*pthat+pow(pthat-sqrt(qthat2),2)),2);
    if(1/z > oozmax) {
      //cout << "Warning: if condition in action" << endl;
      return 0;
    }
    if(1/z < oozmin) {
      //cout << "Warning: if condition in action" << endl;
      return 0;
    }
    double num = -1/z+2*sqrtcoshy/sqrt(z)+qthat2;
    double den = 2*sqrt(qthat2)*pthat;
    double costhetap = num/den;
    if(fabs(costhetap)>1.) {
      cout << "if in action" << endl;
      return 0;
    }
    double thetap = acos(costhetap);
    double z1 = sqrt(z*(1+pthat*pthat))*exp( y);
    double z2 = sqrt(z*(1+pthat*pthat))*exp(-y);
    double phip = phiprime(xi1,xi2,phi);
    double pk2_1 = sqrt(xi2)*pthat*cos(thetap+phip);
    double pk2_2 = sqrt(xi2)*pthat*cos(2*M_PI-thetap+phip);
    //
    double res = Mel2_QQbar(xi1, xi2, z, phi, z1, z2, pk2_1) + Mel2_QQbar(xi1, xi2, z, phi, z1, z2, pk2_2);
    res /= 32*M_PI*M_PI;  // *2pi/2pi/32pi^2
    //cout << res << endl;
    return res / sqrt(den*den-num*num);
  }
  double QQbarSingleQuark::diffXS_Y_pt_v2(double xi, double xi2, double z, double eta, double phip, double xi1) {
    //if(_invM<2*_mQ) return 0; // violates theta function
    double pthat  = _pt/_mQ;
    double qthat2 = xi;
    //double xi1 = xi+xi2-2*sqrt(xi*xi2)*cos(phip);
    double sqrtcoshy = sqrt(1+pthat*pthat)*cosh(eta);
    double oozmax = pow(sqrtcoshy+sqrt(sqrtcoshy*sqrtcoshy-pthat*pthat+pow(pthat+sqrt(qthat2),2)),2);
    double oozmin = pow(sqrtcoshy+sqrt(sqrtcoshy*sqrtcoshy-pthat*pthat+pow(pthat-sqrt(qthat2),2)),2);
    if(1/z > oozmax) {
      cout << "Warning: if condition in action - 1 --   " << 1/z << " > " << oozmax << "   difference: " << 1/z-oozmax << endl;
      z = 1/oozmax + 1e-15;
      //return 0;
    }
    if(1/z < oozmin) {
      cout << "Warning: if condition in action - 2 --   " << 1/z << " < " << oozmin << "   difference: " << 1/z-oozmin << endl;
      //cout << setw(12) << xi << setw(14) << xi2 << setw(14) << z << setw(14) << eta << setw(14) << phip << endl;
      z = 1/oozmin - 1e-15;
      //return 0;
    }
    double num = -1/z+2*sqrtcoshy/sqrt(z)+qthat2;
    double den = 2*sqrt(qthat2)*pthat;
    if(fabs(num)>fabs(den)) {
      cout << "|numerator| grater than |denominator|: " << num << " > " << den << ", difference: " << fabs(num)-fabs(den) << endl;
      num = num/fabs(num)*fabs(den);
    }
    double costhetap = num/den;
    if((fabs(costhetap)-1.)>1e-9) {
      cout << "if in action: |cos(thetap)|-1 = " << fabs(costhetap)-1 << endl;
      //return 0;
    }
    double thetap = myacos(costhetap);
    double z1 = sqrt(z*(1+pthat*pthat))*exp( eta);
    double z2 = sqrt(z*(1+pthat*pthat))*exp(-eta);
    double phi = phinonprime(xi,xi2,phip);
    //double phiptest = phiprime(xi1,xi2,phi);
    //if(fabs(phiptest-phip)>1e-5) cout << "phi' test failed: " << phip << "  " << phiptest << endl;
    double pk2_1 = sqrt(xi2)*pthat*cos(thetap+phip);
    double pk2_2 = sqrt(xi2)*pthat*cos(-thetap+phip);
    //double pk1_1 = sqrt(xi1)*pthat*cos(thetap+phip-phi);
    //double pk1_2 = sqrt(xi1)*pthat*cos(-thetap+phip-phi);
    //double zero = xi+(z1+z2-1)/z-2*pk1_1-2*pk2_2;
    //double zero = 2*sqrt(xi)*pthat*cos(thetap)-2*pk1_1-2*pk2_2;
    //double zero = 2*sqrt(xi)*pthat*cos(thetap) -xi-(z1+z2-1)/z;
    //if(fabs(zero)>1e-7) cout << "zero = "<< zero << endl;
    //
    //double res = Mel2_QQbar(xi1, xi2, z, phi, z1, z2, pk2_1, pk1_1) + Mel2_QQbar(xi1, xi2, z, phi, z1, z2, pk2_2, pk1_2);
    double res = Mel2_QQbar(xi1, xi2, z, phi, z1, z2, pk2_1) + Mel2_QQbar(xi1, xi2, z, phi, z1, z2, pk2_2);
    res /= 32*M_PI*M_PI * 2*M_PI;  // /2pi/32pi^2
    //cout << res << endl;
    if(res != res) cout << "Noooo.... nan" << endl;
    double norm = sqrt(den*den-num*num);
    if(norm != norm) cout << "Yet another nan: num="<<num<<", den="<<den<<", num-den="<< num-den << endl;
    //cout << xi2 << "  " << xi << "  " << z << "  " << eta << "  " << 1/pthat << "  " << thetap << "  " << z1 << endl;
    return res / norm;
  }
  double QQbarSingleQuark::diffXS_Y_pt_v3(double xi1, double xi2, double z, double eta, double phi) {
    //if(_invM<2*_mQ) return 0; // violates theta function
    double pthat  = _pt/_mQ;
    double qthat2 = xi1+xi2+2*sqrt(xi1*xi2)*cos(phi);
    double sqrtcoshy = sqrt(1+pthat*pthat)*cosh(eta);
    double oozmax = pow(sqrtcoshy+sqrt(sqrtcoshy*sqrtcoshy-pthat*pthat+pow(pthat+sqrt(qthat2),2)),2);
    double oozmin = pow(sqrtcoshy+sqrt(sqrtcoshy*sqrtcoshy-pthat*pthat+pow(pthat-sqrt(qthat2),2)),2);
    if(1/z > oozmax) {
      cout << "Warning: if condition in action - 1 --   " << 1/z << " > " << oozmax << "   difference: " << 1/z-oozmax << endl;
      z = 1/oozmax + 1e-15;
      //return 0;
    }
    if(1/z < oozmin) {
      cout << "Warning: if condition in action - 2 --   " << 1/z << " < " << oozmin << "   difference: " << 1/z-oozmin << endl;
      //cout << setw(12) << xi << setw(14) << xi2 << setw(14) << z << setw(14) << eta << setw(14) << phip << endl;
      z = 1/oozmin - 1e-15;
      //return 0;
    }
    double num = -1/z+2*sqrtcoshy/sqrt(z)+qthat2;
    double den = 2*sqrt(qthat2)*pthat;
    if(fabs(num)>fabs(den)) {
      cout << "|numerator| grater than |denominator|: " << num << " > " << den << ", difference: " << fabs(num)-fabs(den) << endl;
      num = num/fabs(num)*fabs(den);
    }
    double costhetap = num/den;
    if((fabs(costhetap)-1.)>1e-9) {
      cout << "if in action: |cos(thetap)|-1 = " << fabs(costhetap)-1 << endl;
      //return 0;
    }
    double thetap = acos(costhetap);
    double z1 = sqrt(z*(1+pthat*pthat))*exp( eta);
    double z2 = sqrt(z*(1+pthat*pthat))*exp(-eta);
    double phip = phiprime(xi1,xi2,phi);
    double pk2_1 = sqrt(xi2)*pthat*cos( thetap+phip);
    double pk2_2 = sqrt(xi2)*pthat*cos(-thetap+phip);
    //
    double res = Mel2_QQbar(xi1, xi2, z, phi, z1, z2, pk2_1) + Mel2_QQbar(xi1, xi2, z, phi, z1, z2, pk2_2);
    res /= 32*M_PI*M_PI * 2*M_PI;  // /2pi/32pi^2
    //cout << res << endl;
    if(res != res) cout << "Noooo.... nan" << endl;
    double norm = sqrt(den*den-num*num);
    if(norm != norm) cout << "Yet another nan: num="<<num<<", den="<<den<<", num-den="<< num-den << endl;
    //cout << xi2 << "  " << xi << "  " << z << "  " << eta << "  " << 1/pthat << "  " << thetap << "  " << z1 << endl;
    return res / norm;
  }
  double QQbarSingleQuark::diffXS_Y_pt_Aux(double xi, double z, double eta) {
    //if(_invM<2*_mQ) return 0; // violates theta function
    double pthat  = _pt/_mQ;
    double B = sqrt(1+pthat*pthat)*cosh(eta);
    //
    double stuff = pthat*pthat + 1/z - 2*B/sqrt(z);
    double sqrtstuff = sqrt(stuff);
    //if(stuff<0 && stuff>-1e-10) sqrtstuff = 0;
    if(sqrtstuff != sqrtstuff) cout << "sqrt(stuff) is NaN again... " << stuff << endl;
    //cout << pthat*pthat + 1/z - sqrt(1+pthat*pthat)*(exp(eta)+exp(-eta))/sqrt(z) << endl;
    double sqrtximax = sqrtstuff + pthat; //min( sqrtstuff + pthat, sqrt(1./z-4));
    double sqrtximin = sqrtstuff - pthat;
    if(xi > sqrtximax*sqrtximax) {
      cout << "Warning: if condition in action - 1 --   " << xi << " > " << sqrtximax*sqrtximax << "   difference: " << xi-sqrtximax*sqrtximax << endl;
      //return 0;
    }
    if(xi < sqrtximin*sqrtximin) {
      cout << "Warning: if condition in action - 2 --   " << xi << " < " << sqrtximin*sqrtximin << "   difference: " << xi-sqrtximin*sqrtximin << endl;
      //return 0;
    }
    //
    /*
    double oozmax = pow(B+sqrt(B*B-pthat*pthat+pow(pthat+sqrt(xi),2)),2);
    double oozmin = pow(B+sqrt(B*B-pthat*pthat+pow(pthat-sqrt(xi),2)),2);
    if(1/z > oozmax) {
      if(fabs(1/z-oozmax)>1e-20) cout << "Warning: if condition in action - 1 --   " << 1/z << " > " << oozmax << "   difference: " << 1/z-oozmax << endl;
      z = 1/oozmax;// + 1e-15;
      //return 0;
    }
    if(1/z < oozmin) {
      if(fabs(1/z-oozmin)>1e-20) cout << "Warning: if condition in action - 2 --   " << 1/z << " < " << oozmin << "   difference: " << 1/z-oozmin << endl;
      //cout << setw(12) << xi << setw(14) << xi2 << setw(14) << z << setw(14) << eta << setw(14) << phip << endl;
      z = 1/oozmin - 1e-15;
      //return 0;
    }
    */
    double num = stuff - pthat*pthat - xi; // 1/z-2*B/sqrt(z)-xi;
    double num2 = num*num;
    double den2 = 4*xi*pthat*pthat;
    if(num2>den2) {
      cout << "num2="<<num2<<" > den2="<<den2<<", num2-den2="<< num2-den2 << "  xi="<<xi << "  z="<<z << "  eta="<<eta << "  ximin="<<sqrtximin*sqrtximin << "  ximax="<<sqrtximax*sqrtximax << endl;
      //den2 = num2 * (1+1e-10);
      num2 = den2 / (1+1e-10);
    }
    double z1 = sqrt(z*(1+pthat*pthat))*exp( eta);
    double z2 = sqrt(z*(1+pthat*pthat))*exp(-eta);
    //
    double res = 2* Mel2_QQbar_partialonshell(xi, z, z1, z2, pthat*pthat);  // the factor of 2 comes from integrating over the delta function in theta'
    res /= 32*M_PI*M_PI;
    if(res != res) cout << "Noooo.... nan" << endl;
    double norm = sqrt(den2-num2);
    if(norm != norm) cout << "Yet another nan: num2="<<num2<<", den2="<<den2<<", num2-den2="<< num2-den2 << endl;
    //cout << xi2 << "  " << xi << "  " << z << "  " << eta << "  " << 1/pthat << "  " << thetap << "  " << z1 << endl;
    return res / norm;
  }

  double QQbarSingleQuark::diffXS_Y_pt_Aux_v1(double xi, double z, double eta) {
    //if(_invM<2*_mQ) return 0; // violates theta function
    double pthat  = _pt/_mQ;
    //
    if(1/z < (1+pthat*pthat)*exp(2*fabs(eta))) {
      return 0;
    }
    if(1/z < 4+xi) {
      return 0;
    }
    //
    double B = sqrt(1+pthat*pthat)*cosh(eta);
    double num = 1/z-2*B/sqrt(z)-xi;
    double num2 = num*num;
    double den2 = 4*xi*pthat*pthat;
    if(num2>den2) {
      return 0;
    }
    double z1 = sqrt(z*(1+pthat*pthat))*exp( eta);
    double z2 = sqrt(z*(1+pthat*pthat))*exp(-eta);
    //
    double res = 2* Mel2_QQbar_partialonshell(xi, z, z1, z2, pthat*pthat);  // the factor of 2 comes from integrating over the delta function in theta'
    res /= 32*M_PI*M_PI;
    if(res != res) cout << "Noooo.... nan" << endl;
    double norm = sqrt(den2-num2);
    if(norm != norm) cout << "Yet another nan: num2="<<num2<<", den2="<<den2<<", num2-den2="<< num2-den2 << endl;
    return res / norm;
  }

  double QQbarSingleQuark::diffXS_Y_pt_Aux_v2(double xi, double z, double eta) {
    //if(_invM<2*_mQ) return 0; // violates theta function
    double pthat  = _pt/_mQ;
    //
    if(1/z < (1+pthat*pthat)*exp(2*fabs(eta))) {
      return 0;
    }
    if(1/z < 4+xi) {
      return 0;
    }
    double z1 = sqrt(z*(1+pthat*pthat))*exp( eta);
    double z2 = sqrt(z*(1+pthat*pthat))*exp(-eta);
    //
    double res = z * Mel2_QQbar_partialonshell(xi, z, z1, z2, pthat*pthat);
    res /= 32*M_PI*M_PI;
    if(res != res) cout << "Noooo.... nan" << endl;
    return res;
  }
  double QQbarSingleQuark::diffXS_Y_pt_Aux_v2_subplus(double z, double eta) {
    //if(_invM<2*_mQ) return 0; // violates theta function
    double pthat  = _pt/_mQ;
    //
    if(1/z < (1+pthat*pthat)*exp(2*fabs(eta))) {
      return 0;
    }
    if(1/z < 4) {
      return 0;
    }
    double z1 = sqrt(z*(1+pthat*pthat))*exp( eta);
    //
    //double res = z * Mel2_QQbar_partialonshell(xi, z, z1, z2, pthat*pthat);
    double res = z * Mel2_QQbar_onshell(z, z1, pthat*pthat);
    res /= 32*M_PI*M_PI;
    if(res != res) cout << "Noooo.... nan" << endl;
    return res;
  }
  double QQbarSingleQuark::diffXS_Y_pt_LO(double z, double eta, double pthat2) {
    //if(_invM<2*_mQ) return 0; // violates theta function
    //double pthat  = _pt/_mQ;
    //
    if(1/z < (1+pthat2)*exp(2*fabs(eta))) {
      return 0;
    }
    if(1/z < 4) {
      return 0;
    }
    double z1 = sqrt(z*(1+pthat2))*exp( eta);
    //
    double res = 2*z * Mel2_QQbar_onshell(z, z1, pthat2)  *2*M_PI;  // 2pi comes from average in theta' that had to be integral in theta'
    res /= 32*M_PI*M_PI;
    if(res != res) cout << "Noooo.... nan   z=" << z << "  eta=" << eta << endl;
    return res;
  }
  double QQbarSingleQuark::diffXS_Y_pt_LO_qqbarchannel(double z, double eta, double pthat2) {
    //if(_invM<2*_mQ) return 0; // violates theta function
    //double pthat  = _pt/_mQ;
    //
    if(1/z < (1+pthat2)*exp(2*fabs(eta))) {
      return 0;
    }
    if(1/z < 4) {
      return 0;
    }
    double res = 2*z * Mel2_QQbar_onshell_qqbarchannel(z, pthat2)  *2*M_PI;  // 2pi comes from average in theta' that had to be integral in theta'
    res /= 32*M_PI*M_PI;
    if(res != res) cout << "Noooo.... nan   z=" << z << "  eta=" << eta << "   qqbar" << endl;
    return res;
  }


  /*
  double QQbarSingleQuark::diffXS_Y_pt_Aux(double intvar1, double intvar2) {
    //if(_invM<2*_mQ) return 0; // violates theta function
    double qthat2 = pow(_pt/_invM,2);
    if(1/_x < 1+qthat2) {
      cout << "Warning: if condition in action" << endl;
      return 0;
    }
    double omega = 2*M_PI*intvar1;
    double beta2 =   M_PI*intvar2;
    //
    double C = Coff(qthat2, omega, beta2) * 2*M_PI*M_PI;
    return C /(1+qthat2);
  }


  // single differential Y
  double QQbarSingleQuark::intphi(double intvar, double xi1, double xi2, double &jac) {
    double cosphibarmax = (1/_x-1-xi1-xi2)/2/sqrt(xi1*xi2);
    double phibar = 0;
    if(cosphibarmax< 1) phibar = acos(cosphibarmax);
    if(cosphibarmax<-1) phibar = M_PI;
    //cout << phibar << endl;
    jac = 2*(M_PI-phibar);
    double phi = phibar + jac*intvar;
    //jac = 2*M_PI;
    //double phi = jac*intvar3;
    return phi;
  }
  double QQbarSingleQuark::diffXS_Y(double xi1, double xi2, double intvar1, double intvar2, double intvar3) {
    //if(_invM<2*_mQ) return 0; // violates theta function
    double jac;
    double omega = 2*M_PI*intvar1;
    //double beta1 =   M_PI*intvar1;
    double beta2 =   M_PI*intvar2;
    if(1/_x-1<pow(sqrt(xi1)-sqrt(xi2),2)) return 0;
    double phi   = intphi(intvar3, xi1, xi2, jac);
    double cosphi = cos(phi);
    //
    double qthat2 = xi1+xi2+2*sqrt(xi1*xi2)*cosphi;
    if(1/_x < 1+qthat2) {
      cout << "Warning: if condition in action" << endl;
      return 0;
    }
    double dCdphi = dCoff_dphi(xi1, xi2, phi, omega, beta2) * 2*M_PI*M_PI*jac;
    return dCdphi /(1+qthat2);// /(1+qthat2);
  }
  double QQbarSingleQuark::diffXS_Y_Aux(double intvar1, double intvar2) {
    //if(_invM<2*_mQ) return 0; // violates theta function
    double z = _x*exp(-2*_rapY);
    double qthat2 = 1/z-1;
    if(fabs(_rapY) > -0.5*log(_x)) {
      cout << "Warning: if condition in action" << endl;
      return 0;
    }
    double omega = 2*M_PI*intvar1;
    double beta2 =   M_PI*intvar2;
    //
    double C = Coff(qthat2, omega, beta2) * 2*M_PI*M_PI;
    return C;
  }
  */























  










  /*
  // double QQbarPair::zzdCoff_dphi_v1(double xi1, double xi2, double cosphi, double beta1, double beta2) { // the last two are phase-space integration vars
  //   double qthat2 = xi1+xi2+2*sqrt(xi1*xi2)*cosphi;
  //   double z = 1./(1.+qthat2);
  //   double m2Q2 = _mQ*_mQ/_invM/_invM;
  //   double z1 = 0.5-sqrt(0.25-m2Q2)*cos(beta1);
  //   double z2 = 0.5-sqrt(0.25-m2Q2)*cos(beta2); // in realta` dovrebbe essere un ellisse
  //   //
  //   double pthat2 = z1*z2/z-m2Q2;
  //   double costhetap = ( qthat2-(1-z1-z2)/z ) /sqrt(qthat2*pthat2)/2.;
  //   if(fabs(costhetap)>1.) {
  //     //cout << "if in action  " << costhetap << endl;
  //     return 0;
  //   }
  //   double abssinthetap = sqrt(1-costhetap*costhetap);
  //   double M2 = 0;
  //   //
  //   for(int i=0; i<2; i++) {
  //     double sinthetap = (i==0? 1: -1) * abssinthetap;
  //     double twok2phat = sqrt(pthat2/qthat2) * ( (qthat2-xi1+xi2)*costhetap + sqrt((qthat2+xi1+xi2)*(qthat2+xi1+xi2)-2*(qthat2*qthat2+xi1*xi1+xi2*xi2))*sinthetap );
  //     double twok1phat = sqrt(pthat2/qthat2) * ( (qthat2-xi2+xi1)*costhetap + sqrt((qthat2+xi1+xi2)*(qthat2+xi1+xi2)-2*(qthat2*qthat2+xi1*xi1+xi2*xi2))*sinthetap );
  //     double T = twok1phat -xi1 - z2/z;
  //     double U = twok2phat -xi2 - z1/z;
  //     double Delta = (z1-z2)/z + xi1*z2 - xi2*z1 + (U+xi2-T-xi1)/2.;
  //     double M2ab  =  1/T/U - pow(1+z2*(1-z1)/z/T+z1*(1-z2)/z/U,2)/xi1/xi2;
  //     double M2nab = -1/T/U + (z2-z1)*(1/T-1/U) + 2*z + 2*(0.5+z2*(1-z1)/z/T-Delta)*(0.5+z1*(1-z2)/z/U+Delta)/xi1/xi2;
  //     M2 += ( M2ab/2./CA + M2nab/CF ); // /z/z;
  //     //M2 += M2ab/2./CA; // /z/z;
  //   }
  //   return M2 * (0.25-m2Q2)*sin(beta1)*sin(beta2) /z/sqrt(pthat2*qthat2)/abssinthetap/32./M_PI/M_PI;
  // }


  /////////////////////////////////////////////////////////
  /////// CHECK COMMIT 21/07/2020 at 9.53  ----------------
  
  double zzMelQQ(double xi1, double xi2, double phi, double z1, double z2, double thetap, double m2Q2) {  // checked against mathematica
    double qthat2 = xi1+xi2+2*sqrt(xi1*xi2)*cos(phi);
    //double pthat = ((z1+z2)/z-1)/(2*sqrt(xi1)*cos(thetap)+2*sqrt(xi2)*cos(thetap-phi));
    double pthat = ((z1+z2)*(1+qthat2)-1)/(2*sqrt(xi1)*cos(thetap)+2*sqrt(xi2)*cos(thetap-phi));
    //double z = 1./(1.+qthat2); // true only if m2Q2=0
    double z = z1*z2/(m2Q2+pthat*pthat);
    //double T = z1 * (xi1-xi2+1) - xi1 - twok2Delta - 1;
    //double U = z1 * (xi2-xi1-1) - xi2 + twok2Delta;
    //double twok1p = 2*pthat*sqrt(xi1)*cos(thetap);
    //double U = qthat2 - xi2 - twok1p - (1-z2)/z;
    //double T =        - xi1 + twok1p -    z2 /z;
    double T = xi2 + 2*sqrt(xi2*xi1)*cos(phi) - 2*sqrt(xi2)*pthat*cos(thetap-phi) - (1-z1)/z;
    double U = 2*sqrt(xi2)*pthat*cos(thetap-phi) - xi2 - z1 /z;
    double Delta = (z1-z2)/z + xi1*z2 - xi2*z1 + (U+xi2-T-xi1)/2.;
    double M2ab  =  1/T/U - pow(1+z2*(1-z1)/z/T+z1*(1-z2)/z/U,2)/xi1/xi2;
    double M2nab = -1/T/U + (z2-z1)*(1/T-1/U) + 2*z + 2*(0.5+z2*(1-z1)/z/T-Delta)*(0.5+z1*(1-z2)/z/U+Delta)/xi1/xi2;
    double M2 = ( M2ab/2./CA + M2nab/CF );
    return M2; // /z/z;
  }
  double zzMelQQ0(double xi1, double z1, double z2, double thetap) {  // checked against mathematica, plus internal cross-check of on-shell limit
    double z = 1./(1.+xi1);
    double pthat2 = pow(((z1+z2)/z-1)/(2*sqrt(xi1)*cos(thetap)),2);
    //
    double M2ab  = 1/z1/(1-z1) - 2*z2/(1-z1)*(1+z2/z1-z2*z2/z1/(1-z1)) + 2*(1-z1-z2)*(1-z1-z2)/z1/z1/(1-z1)/(1-z1)/xi1*(z1*z2-pthat2);
    double M2nab = -1/z1/(1-z1) - (1/z1-1/(1-z1))*(z1-z2)/z+2/z
      + 2/z/z/xi1*( -2*z2*z2*(1-z/(1-z1))*xi1 -2*(1-z*z2/(1-z1))*(1-z*(1-z2)/z1)*pthat2
    		    +z2*((1-z/(1-z1))*(1-z*(1-z2)/z1)+1-z*z2/(1-z1))*(xi1-(1-z1-z2)/z));
    double M2 = ( M2ab/2./CA + M2nab/CF );
    return z*z*M2; // /z/z;
  }

  double QQbarPair::zzdCoff_dphi(double xi1, double xi2, double phi, double omega, double beta) { // the last two are phase-space integration vars
    double qthat2 = xi1+xi2+2*sqrt(xi1*xi2)*cos(phi);
    double z = 1./(1.+qthat2);
    double m2Q2 = _mQ*_mQ/_invM/_invM;
    //
    double theta = acos((sqrt(xi1)+sqrt(xi2)*cos(phi))/sqrt(qthat2));
    if(sin(phi)<0) theta = 2*M_PI - theta; // corrects the angle if it's between pi and 2*pi [comes from the fact that sin(theta)\propto sin(phi)]
    //
    double absDeltahat = sqrt(1-4*m2Q2)*sin(beta)/2.;
    //double twok2Delta = 2*sqrt(xi2)*absDeltahat * cos(omega-phi);
    //
    double z1 = 0.5-sqrt(0.25-m2Q2)*cos(beta);
    double z2 = z * ( 1 - (1-qthat2)*z1 + 2*sqrt(qthat2)*absDeltahat*cos(omega-theta) );
    if(sqrt(0.25-m2Q2)-fabs(0.5-z2)<0) { cout << "eeeeeee" << endl; }
    if(z2<0 || z2>1) { cout << "ooooooo" << endl; }
    //
    double pthat2 = z1*z1*qthat2 + absDeltahat*absDeltahat + 2*z1*sqrt(qthat2)*absDeltahat*cos(omega-theta);
    double thetap = acos((z1*sqrt(xi1)+z1*sqrt(xi2)*cos(phi)+absDeltahat*cos(omega))/sqrt(pthat2));
    if(z1*sqrt(xi2)*sin(phi)+absDeltahat*sin(omega)<0) thetap = 2*M_PI - thetap; // corrects the angle if it's between pi and 2*pi
    //
    //double T = z1 * (xi1-xi2+1) - xi1 - twok2Delta - 1;
    //double U = z1 * (xi2-xi1-1) - xi2 + twok2Delta;
    //double Delta = (z1-z2)/z + xi1*z2 - xi2*z1 + (U+xi2-T-xi1)/2.;
    //double M2ab  =  1/T/U - pow(1+z2*(1-z1)/z/T+z1*(1-z2)/z/U,2)/xi1/xi2;
    //double M2nab = -1/T/U + (z2-z1)*(1/T-1/U) + 2*z + 2*(0.5+z2*(1-z1)/z/T-Delta)*(0.5+z1*(1-z2)/z/U+Delta)/xi1/xi2;
    //double M2 = ( M2ab/2./CA + M2nab/CF ); ///z/z;
    //double M2 = M2ab/2./CA; // /z/z;
    double M2 = zzMelQQ(xi1, xi2, phi, z1, z2, thetap, m2Q2);
    return M2 * sqrt(0.25-m2Q2)*sin(beta) /16./M_PI/M_PI;
  }


  double QQbarPair::zzCoff(double xi1, double omega, double beta) { // the last two are phase-space integration vars
    double z = 1./(1.+xi1);
    double m2Q2 = _mQ*_mQ/_invM/_invM;
    //
    double absDeltahat = sqrt(1-4*m2Q2)*sin(beta)/2.;
    double z1 = 0.5-sqrt(0.25-m2Q2)*cos(beta);
    double z2 = z * ( 1 - (1-xi1)*z1 + 2*sqrt(xi1)*absDeltahat*cos(omega) );
    if(sqrt(0.25-m2Q2)-fabs(0.5-z2)<0) { cout << "eeeeeee" << endl; }
    if(z2<0 || z2>1) { cout << "ooooooo" << endl; }
    //
    double pthat2 = z1*z1*xi1 + absDeltahat*absDeltahat + z1*sqrt(xi1)*absDeltahat*cos(omega);
    //
    double M2ab  = 1/z1/(1-z1) - 2*z2/(1-z1)*(1+z2/z1-z2*z2/z1/(1-z1)) + 2*(1-z1-z2)*(1-z1-z2)/z1/z1/(1-z1)/(1-z1)/xi1*(z1*z2-pthat2);
    double M2nab = -1/z1/(1-z1) - (1/z1-1/(1-z1))*(z1-z2)/z+2/z
      + 2/z/z/xi1*( -2*z2*z2*(1-z/(1-z1))*xi1 -2*(1-z*z2/(1-z1))*(1-z*(1-z2)/z1)*pthat2
    		    +z2*((1-z/(1-z1))*(1-z*(1-z2)/z1)+1-z*z2/(1-z1))*(xi1-(1-z1-z2)/z));
    //double M2nab = 1/z1/(1-z1)/xi1*(
    //				    - xi1 + 4*pthat2*((1+xi1)*(-1+z1)+z2)*(-1+z1+xi1*z1+z2)
    //				    + xi1*(1+xi1)*(z1+z2-2*z1*z2+4*z1*(xi1*(-1+z1)+z1)*z2*z2)
    //				    - 2*z2*(-1 + (1 + xi1)* (z1 + z2))* (2* xi1*xi1* (-1 + z1)* z1
    //									 + 2* z1* (-1 + z1 + z2) + xi1* (-1 + 2* z1)* (-1 + 2* z1 + z2))
    //				    );
    double M2 = ( M2ab/2./CA + M2nab/CF );
    //double M2 = M2ab/2./CA;
    return z*z*M2 * sqrt(0.25-m2Q2)*sin(beta) /16./M_PI/M_PI;
  }


  double QQbarPair::zzC(double beta) { // phase-space integration var
    double m2Q2 = _mQ*_mQ/_invM/_invM;
    //
    double z1 = 0.5-sqrt(0.25-m2Q2)*cos(beta);
    double pthat2 = (1-4*m2Q2)/4.*pow(sin(beta),2);
    //
    double par = 1/z1/(1-z1);
    double M2ab  = par - 2 - 4*(par*par*pthat2*pthat2 - par*pthat2);
    //double M2nab = -par + 1/z1-1/(1-z1) + 2*z1/(1-z1) +4/par + 4*(2*par*pthat2-1)*pthat2;
    double M2nab = 4/par - 2 - 8*pthat2 + 8*par*pthat2*pthat2;
    double M2 = ( M2ab/2./CA + M2nab/CF );
    //double M2 = M2ab/2./CA;
    return M2 * sqrt(0.25-m2Q2)*sin(beta) /8./M_PI;
  }

  */


  struct s_xxx { QQbarPair *pair; double d1, d2, d3, d4, d5; };
  double integrandGSLxxx1(double x, void *pars) {
    s_xxx p = *(s_xxx*)pars;
    double xi1 = p.d1;
    double xi2 = p.d2;
    double ome = p.d3;
    double bet = p.d4;
    double phi = 2*M_PI*x;
    double result = p.pair->dCoff_dphi(xi1, xi2, phi, ome, bet) * 2*M_PI;
    //double result = p.pair->dCoff_dphi_0(xi1, phi, ome, bet) * 2*M_PI;
    return result;
  }
  double QQbarPair::TESTzzdCoff_dphi_int(double xi1, double xi2, double omega, double beta) {
    //s_xxx pars = { this, xi1, xi2, z1, z2, theta };
    s_xxx pars = { this, xi1, xi2, omega, beta, 0 };
    gsl_error_handler_t *old_handler = gsl_set_error_handler (NULL);
    gsl_set_error_handler_off();
    double prec = 1e-7;
    int status;
    double res = integration_gauss(integrandGSLxxx1, 0, 1, prec, NULL, &pars, &status);
    //if(status) res = 0;
    gsl_set_error_handler (old_handler);
    return res;
  }
  double integrandGSLxxx2(double x, void *pars) {
    s_xxx p = *(s_xxx*)pars;
    double xi1 = p.d1;
    double bet = p.d4;
    double ome = 2*M_PI*x;
    double result = p.pair->Coff(xi1, ome, bet) * 2*M_PI;
    return result;
  }
  double QQbarPair::TESTzzCoff_int(double xi1, double beta) {
    s_xxx pars = { this, xi1, 0, 0, beta, 0 };
    //gsl_error_handler_t *old_handler = gsl_set_error_handler (NULL);
    //gsl_set_error_handler_off();
    double prec = 1e-7;
    int status;
    double res = integration_gauss(integrandGSLxxx2, 0, 1, prec, NULL, &pars, &status);
    //if(status) res = 0;
    //gsl_set_error_handler (old_handler);
    return res;
  }






};
